### Module containing functions to compute some baselines for BGL prediction, as well
### as two of the evaluation methods used in the paper (MSE and ternary)

### Sadegh M, Summer 2016

from sklearn import svm
from sklearn.preprocessing import PolynomialFeatures
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import Pipeline
import numpy as np

def MSE(pred, target):
    """
    computes mean-square-error of a prediction and target vector,
    it is done element by element so that any array-like input can be taken.
    TAKES:
        pred: an array-like prediction vector
        target: an array-like target vector, the same size as pred
    RETURNS:
        mean-square-error between pred and target
    """
    N= len(pred)
    if N < len(target):
        print("Warning: passed prediction vector shorter than target vector")

    s = 0
    for i in range(N):
        s += (pred[i] - target[i]) ** 2

    return (s / N)

def ternary(t0, targets, pred, sameTolerance=10):
    """
    Computes the ternary cost value (see the ICMLA13 paper for description).
    TAKES:
        t0, targets, pred: iterables containing the values at t0, t0+prediction horizon
            and the predicted values respectively
        sameTolerance: values within this range of t0 will be considered no change
    OUTPUTS:
        the cost value as the sum of all costs
    """
    cost = 0

    for i in range(len(t0)):
        a = p = 0

        if targets[i]-t0[i] > sameTolerance:
            a = 1
        elif targets[i]-t0[i] < sameTolerance:
            a = -1

        if pred[i]-t0[i] > sameTolerance:
            p = 1
        elif pred[i]-t0[i] < sameTolerance:
            p = -1

        cost += abs(a - p)

    return cost
    
def bgl_time_delay(pred, target, cutoff_val, comparator):
    """
    Computes a list of delays in prediction of hypo/hyper-glycemia events given
    the cutoff values provided. To get hypo delays, provide hypo as cutoff_val
    and a less-than function as comparator.
    TAKES:
        pred and target: a numpy list of prediction and target values.
        comparator: a function with two args that compares prediction
            and target values (first argument) with the cutoff.
    RETURNS:
        a list containing delays for all the detected regions in 'target'
    """
    assert(len(pred)==len(target))
    predCuts = comparator(pred, cutoff_val)
    targetCuts = comparator(target, cutoff_val)

    res = []
    inRegion = False
    doneRegion = False
    delay = 0

    for i in range(len(targetCuts)):
        if targetCuts[i]:
            if not inRegion:
                inRegion = True
                doneRegion = False
        else:
            if inRegion and not doneRegion:
                res.append(delay)
                delay = 0
            inRegion = False
            
        
        if inRegion and not doneRegion:
            if predCuts[i]:
                res.append(delay)
                delay = 0
                doneRegion = True
            else:
                delay += 1

    return np.asarray(res)


def t0_baseline(data, dataPart='test', bglInd=0):
    """
    TAKES:
      data: a dictionary in the form {pid:{key:value}}.
      dataPart: part of data in the data dictionary to use for computations
      bglInd: the index of BGL values in the data
    RETURNS:
      a list of RMSE values for every patient in data
    """
    res = []
    for pid in data:
      t0 = linear_extrap(data[pid][dataPart][:, :, bglInd], -1, 1)
      res.append(MSE(t0, data[pid][dataPart+'Labels'][:, bglInd])**0.5)
    
    return res
    

def linear_extrap(npData, dt, L=2):
    """
    returns the extrapolation prediction based on a slope computed from the average
    of a given number of previous slopes in the sequence.
    TAKES:
        npData: a N-by-T numpy matrix of N sequences of length T
        dt: an int indicating the units of time into the future to predict
        L: the index deepest in prev history; L=2 means the previous one t-1, L=1 means a t0 baseline.
    RETURNS:
        the predicted labels at t0+dt as a N-by-1 numpy vector
    """
    N, T = npData.shape

    if T < L or L < 1:
        raise ValueError

    if L > 1:
      m = (npData[:, -1] - npData[:, -L]) / (L - 1)
    else:
      m = 0

    return npData[:, -1] + dt*m


def SVR(trainData, trainLabels, testData, kernel='rbf', C=1000, gamma=10):
    """
    uses the support vector regression from scikit-learn to train a model and returns
    the prediction vector on the test data.
    TAKES:
        trainData: a N-by-T numpy matrix of N sequences of length T
        trainLabels: a vector of size N
        testData: same as trainData
        C, gamma: SVM parameters, larger gamma equals more fitting power
    RETURNS:
        the prediction vector on the test data
    """

    model = svm.SVR(kernel=kernel, C=C, gamma=gamma)
    model.fit(trainData, trainLabels)

    return model.predict(testData)


def SVR_tune(trainData, trainLabels, testData, testLabels=[], kernel='rbf', C=[1000], gamma=[10]):
    """
    uses the support vector regression from scikit-learn to train a model on given
    training data, test it on given test data, for a range of C and gamma values.
    TAKES:
        trainData: a N-by-T numpy matrix of N sequences of length T
        trainLabels: a vector of size N
        testData: same as trainData
        testLabels: same as testLabels
        C, gamma: a list of values for the parameters
    RETURNS:
        MSError on the training and test data for all the parameters, as a pair
        of two matrices, first one being the training performance.
    """
    trnRes = np.zeros([len(C), len(gamma)])
    tstRes = np.zeros([len(C), len(gamma)])

    for i in range(len(C)):
        for j in range(len(gamma)):
            model = svm.SVR(kernel=kernel, C=C[i], gamma=gamma[j])
            model.fit(trainData, trainLabels)
            trnRes[i, j] = MSE( model.predict(trainData), trainLabels )
            tstRes[i, j] = MSE( model.predict(testData), testLabels )

    return trnRes, tstRes


def poly_regression(trainData, trainLabels, testData, degree=2):
    """
    uses polynomial regression from scikit-learn to train a model and returns
    the prediction vector on the test data.
    TAKES:
        trainData: a N-by-T numpy matrix of N sequences of length T
        trainLabels: a vector of size N
        testData: same as trainData
    RETURNS:
        the prediction vector on the test data
    """
    model = Pipeline([('poly', PolynomialFeatures(degree=degree)),
                      ('linear', LinearRegression(fit_intercept=False))])
    model = model.fit(trainData, trainLabels)

    return model.predict(testData)

def arima(npData, dt, p, d, q, verbose=1):
    """
    uses the ARIMA class from statsmodel.tsa.arima_model to fit N sequences
    of length T and produce a prediction for each at time dt after the last.
    TAKES:
        npData: a N-by-T numpy matrix of N sequences of length T
        dt: a float indicating the units of time into the future to predict
        p, d, q: ARIMA parameters
    RETURNS:
        a vector of size N, predictions for dt time units after the last of each sequence
    """
    from statsmodels.tsa.arima_model import ARIMA

    N, T = npData.shape
    output = []

    for n in range(N):
        if verbose:
            print("--> fitting ARIMA on sequence {0} of {1}".format(n, N))
        model = ARIMA(npData[n, :], [p, d, q])
        fittedModel = model.fit()
        output.append( fittedModel.predict(T, T+dt-1)[-1] )

    return output

