# -*- coding: utf-8 -*-
"""
Contains definitions of neural network models that can be used to train, evaluate and test.

Created on Sun May 13 2018
@author: Sadegh M
"""

import keras
import tensorflow as tf
import numpy as np
import warnings


class EarlyStoppingChkpnt(keras.callbacks.EarlyStopping):
    """
    A modification of Keras's earlystopping class that enables saving the best weight in variable best_weights.
    """

    def __init__(self, *args, **kwargs):
        keras.callbacks.EarlyStopping.__init__(self, *args, **kwargs)
        self.best_weights = None
        self.best = None

    def on_train_begin(self, logs=None):
        # Allow instances to be re-used
        self.wait = 0
        self.stopped_epoch = 0
        self.best = np.Inf if self.monitor_op == np.less else -np.Inf
        self.best_weights = self.model.get_weights()

    def on_epoch_end(self, epoch, logs=None):
        current = logs.get(self.monitor)
        if current is None:
            warnings.warn(
                'Early stopping conditioned on metric `%s` '
                'which is not available. Available metrics are: %s' %
                (self.monitor, ','.join(list(logs.keys()))), RuntimeWarning
            )
            return
        if self.monitor_op(current - self.min_delta, self.best):
            self.best = current
            self.best_weights = self.model.get_weights()
            self.wait = 0
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_epoch = epoch
                self.model.stop_training = True


########################################################################################################################


class NnModel(object):
    """
    Parent class for NN models
    """

    def __init__(self, args, dims):
        self.D = dims['D']  # dimensionality of the input, can be a tuple if multiple inputs
        self.optimizer = args.optimizer
        self.internal_model = None
        self.loss = None  # can be a tuple if multiple losses (corresponding to multiple output)

    @staticmethod
    def config_tf(gpu_id, gpu_mem):
        config = tf.ConfigProto()
        config.gpu_options.allow_growth = False
        config.gpu_options.visible_device_list = str(gpu_id)
        config.gpu_options.per_process_gpu_memory_fraction = gpu_mem
        return config

########################################################################################################################


class KerasModel(NnModel):
    """
    Parent class for Keras models.
    """

    def __init__(self, args, dims):
        NnModel.__init__(self, args, dims)

        if keras.backend.backend() == 'tensorflow':
            print("Keras is using tensorflow version ", tf.__version__)
            config = self.config_tf(args.gpu_id, args.gpu_mem)
            keras.backend.tensorflow_backend.set_session(tf.Session(config=config))

    def compile(self, init_lrate, lrate_decay=0.0, metrics=None):
        if self.optimizer == 'sgd':
            optimizer_func = keras.optimizers.SGD
        elif self.optimizer == 'rmsprop':
            optimizer_func = keras.optimizers.RMSprop
        elif self.optimizer == 'adam':
            optimizer_func = keras.optimizers.Adam
        elif self.optimizer == 'nadam':
            optimizer_func = keras.optimizers.Nadam
        else:
            raise Exception("Unknown optimizer '{}'".format(self.optimizer))

        self.internal_model.compile(loss=self.loss,
                                    optimizer=optimizer_func(lr=init_lrate, decay=lrate_decay),
                                    metrics=metrics)

    def fit(self, x, y, val_portion=0.0, val_data=None, batch_size=32, early_stop=None,
            callbacks=None, max_epochs=1):

        if early_stop is not None:
            early_stopper = EarlyStoppingChkpnt(monitor='val_loss', patience=early_stop)
            callbacks = callbacks or []
            callbacks += [early_stopper]

        fit_res = self.internal_model.fit(x, y,
                                          epochs=max_epochs,
                                          validation_split=val_portion,
                                          validation_data=val_data,
                                          shuffle=True,
                                          batch_size=batch_size,
                                          callbacks=callbacks)
        if early_stop is not None:
            print("Restoring the best weights with loss ", early_stopper.best)
            self.internal_model.set_weights(early_stopper.best_weights)

        return fit_res

    def predict(self, x, batch_size=32):
        return self.internal_model.predict(x, batch_size=batch_size)

    def evaluate(self, x, y, batch_size=32):
        return self.internal_model.evaluate(x, y, batch_size=batch_size)

########################################################################################################################


class KerasSimpleRnn(KerasModel):
    """
    Keras RNN model with a linear neuron at the output.
    """

    def __init__(self, args, dims):
        KerasModel.__init__(self, args, dims)
        self.seq_len = dims['seq_len']
        self.num_layers = args.num_layers
        self.layer_size = args.layer_size
        self.kernel_activation = args.kernel_activation
        self.kernel_initializer = args.kernel_initializer
        self.output_dropout = args.output_dropout
        self.loss = 'mse'

    def generate(self):
        model = keras.models.Sequential()
        for i in range(self.num_layers):
            if i == self.num_layers - 1:
                _output_seq = False
            else:
                _output_seq = True

            model.add(keras.layers.recurrent.LSTM(units=self.layer_size,
                                                  input_shape=(self.seq_len, self.D),
                                                  activation=self.kernel_activation,
                                                  recurrent_activation='sigmoid',
                                                  implementation=2,
                                                  kernel_initializer=self.kernel_initializer,
                                                  dropout=self.output_dropout,
                                                  return_sequences=_output_seq))

        model.add(keras.layers.Dense(1, activation='linear'))

        self.internal_model = model
        return self.internal_model
