# Contains definitions for NN model classes in Keras, inheriting from classes in nn_models_base.py

import keras
import nn_models_base as nn
from keras import layers as krLayers


class DoubleLstm(nn.KerasSimpleRnn):
    """
    A double LSTM model to handle the what-if scenario.
    """

    def __init__(self, args, dims, mode='basic',
                 double_output=False, second_layer_size=None, whatif_inds=None):
        nn.KerasSimpleRnn.__init__(self, args, dims)
        self.second_layer_size = second_layer_size or self.layer_size
        self.pred_horizon = dims['pred_horizon']
        self.mode = mode
        self.double_output = double_output
        self.whatif_inds = whatif_inds

    def generate(self):
        last_lay = inp1 = krLayers.Input(shape=(self.seq_len, self.D), name='input_with_bgl')

        if self.whatif_inds is None:
            inp2_size = self.D - 1
        else:
            inp2_size = len(self.whatif_inds)

        inp2 = krLayers.Input(shape=(self.pred_horizon, inp2_size), name='input_without_bgl')

        for i in range(self.num_layers):
            if i == self.num_layers - 1:
                _output_seq = False
            else:
                _output_seq = True

            last_lay, last_h, last_c = krLayers.recurrent.LSTM(units=self.layer_size,
                                                               activation=self.kernel_activation,
                                                               recurrent_activation='sigmoid',
                                                               implementation=2,
                                                               kernel_initializer=self.kernel_initializer,
                                                               dropout=self.output_dropout,
                                                               return_state=True,
                                                               return_sequences=_output_seq)(last_lay)

        mapped_h = krLayers.Dense(units=self.layer_size, activation='tanh')(last_h)
        mapped_c = krLayers.Dense(units=self.layer_size, activation='tanh')(last_c)

        lay2 = krLayers.recurrent.LSTM(units=self.second_layer_size,
                                       activation=self.kernel_activation,
                                       recurrent_activation='sigmoid',
                                       implementation=2,
                                       kernel_initializer=self.kernel_initializer,
                                       dropout=self.output_dropout)(inp2, initial_state=[mapped_h, mapped_c])

        if self.mode == 'basic':
            _feed_to_last_layer = lay2
        elif self.mode == 'feed_last_h_to_predictor':
            _feed_to_last_layer = krLayers.concatenate([mapped_h, lay2])
        else:
            raise Exception("The provided mode is not recognized: ", self.mode)

        outputs = [krLayers.Dense(1, activation='linear')(_feed_to_last_layer)]

        if self.double_output:
            outputs.append(krLayers.Dense(1, activation='linear')(last_h))

        self.internal_model = keras.models.Model(inputs=[inp1, inp2], outputs=outputs)
        return self.internal_model

    def fit(self, x, y, val_data=None, *args, **kwargs):
        """
        x is assumed to have seq_len+pred_horizon elements, all with available BGL. The BGL will not be used after
        t0, but may be used for labeling. Only features listed in the whatif_inds tuple will be used as whatif information.
        """
        q = self.seq_len

        if self.whatif_inds is None:
            wii = slice(1, None)
        else:
            wii = list(self.whatif_inds)

        x = [x[:, :q, :], x[:, q:, wii]]

        if self.double_output:
            y = [y, y]

        if val_data is not None:
            x_val, y_val = val_data
            x_val = [x_val[:, :q, :], x_val[:, q:, wii]]
            if self.double_output:
                y_val = [y_val, y_val]
            val_data = (x_val, y_val)

        return super(DoubleLstm, self).fit(x, y, val_data=val_data, *args, **kwargs)

    def predict(self, x, batch_size=32):
        if self.whatif_inds is None:
            wii = slice(1, None)
        else:
            wii = list(self.whatif_inds)

        x = [x[:, :self.seq_len, :], x[:, self.seq_len:, wii]]

        return self.internal_model.predict(x, batch_size=batch_size)

    def evaluate(self, x, y, batch_size=32, return_output1_only=True):
        if self.double_output:
            y = [y, y]

        if self.whatif_inds is None:
            wii = slice(1, None)
        else:
            wii = list(self.whatif_inds)

        x = [x[:, :self.seq_len, :], x[:, self.seq_len:, wii]]

        res = self.internal_model.evaluate(x, y, batch_size=batch_size)

        if self.double_output and return_output1_only:
            res = res[1]

        return res

########################################################################################################################
