# Contains definitions for model classes and trainer functions. A model object
# establishes the graph between given input tensors and an output loss tensor.
# Trainer functions take that loss function and output a trainer op.

import tensorflow as tf
from patient_data import batch_generator


################################################################################


class Model(object):
  """
  Base class for TF models, which are objects mainly for defining an input/output graph.
  """
  def __init__(self, args, input_ts):
    self.output_dropout = args.output_dropout
    self.layer_size     = args.layer_size
    self.activation     = args.kernel_activation
    self.kernel_initializer = args.kernel_initializer
    self.input_ts       = input_ts
    self.loss           = None

    self.output_dropout_ts  = tf.placeholder(tf.float32)




class RNNTSAModel(Model):
  """
  An RNN TSA model is an RNN model that outputs a real value, i.e. a regression model.
  """
  def __init__(self, args, input_ts):
    Model.__init__(self, args, input_ts)
    self.cell_type     = args.cell_type

    N = tf.shape(self.input_ts)[0]
    T = tf.shape(self.input_ts)[1]

    if self.cell_type == 'rnn':
      cell_fn = tf.nn.rnn_cell.BasicRNNCell
    elif self.cell_type == 'gru':
      cell_fn = tf.nn.rnn_cell.GRUCell
    elif self.cell_type == 'lstm':
      cell_fn = tf.nn.rnn_cell.LSTMCell
    else:
      raise Exception("Node type not supported: {}".format(self.cell_type))

    if self.activation == 'tanh':
      activation_fn = tf.nn.tanh
    elif self.activation == 'relu':
      activation_fn = tf.nn.relu
    elif self.activation == 'softplus':
      activation_fn = tf.nn.softplus
    elif self.activation == 'sigmoid':
      activation_fn = tf.nn.sigmoid
    else:
      raise Exception("Activation not supported: {}".format(self.activation))


    if self.kernel_initializer == 'glorot_normal':
      _initializer = tf.contrib.layers.xavier_initializer(uniform=False, seed=None, dtype=tf.float32)
    elif self.kernel_initializer == 'glorot_uniform':
      _initializer = tf.contrib.layers.xavier_initializer(uniform=True, seed=None, dtype=tf.float32)
    else:
      raise Exception("Kernel initializer not supported: {}".format(self.kernel_initializer))

    _layer_list = []
    for _ in range(args.num_layers):
      _layer = cell_fn(self.layer_size, forget_bias=1.0, activation=activation_fn,
                       initializer=_initializer)
      if self.output_dropout > 0.0:
        _layer = tf.nn.rnn_cell.DropoutWrapper(_layer, output_keep_prob=1.0-self.output_dropout_ts)

      _layer_list.append(_layer)

    layers = tf.nn.rnn_cell.MultiRNNCell(_layer_list)

    self.initial_states = layers.zero_state(N, tf.float32)

    rnn_outputs, self.rnn_final_state = tf.nn.dynamic_rnn(layers, self.input_ts,
                                                initial_state=self.initial_states,
                                                dtype=tf.float32)

    ## below are two different ways of achieving the same thing, one should be commented out
    #indices = tf.stack([tf.range(N), (T-1)*tf.ones([N], tf.int32)], axis=1)
    #rnn_last_output = tf.gather_nd(rnn_outputs, indices) # NxRNNSIZE

    rnn_last_output = tf.transpose(rnn_outputs, [1, 0, 2])[-1]

    self.reg_output = tf.layers.dense(inputs=rnn_last_output, units=1, activation=None,
                                kernel_initializer=_initializer,
                                kernel_regularizer=None) # Nx1




class FFModel(Model):
  """
  Represents a fully connected NN, aka multilayered perceptron (MLP) or just a deep net.
  """
  def __init__(self, args, input_ts, vocab_size):
    Model.__init__(self, args, input_ts)

    if self.activation == 'tanh':
      activation_fn = tf.nn.tanh
    elif self.activation == 'relu':
      activation_fn = tf.nn.relu
    elif self.activation == 'softplus':
      activation_fn = tf.nn.softplus
    elif self.activation == 'sigmoid':
      activation_fn = tf.nn.sigmoid
    else:
      raise Exception("Activation not supported: {}".format(self.activation))

    _initializer = tf.contrib.layers.xavier_initializer(uniform=True, seed=None, dtype=tf.float32)
    _activations = self.input_ts
    for _ in range(args.num_layers):
      _activations = tf.layers.dense(inputs=_activations, units=self.layer_size,
                                activation=activation_fn, kernel_initializer=_initializer,
                                kernel_regularizer=None)
      if self.output_dropout > 0.0:
        _activations = tf.layers.dropout(inputs=_activations, rate=self.output_dropout,
                                         training=tf.cond(self.output_dropout_ts > 0.0, lambda: True, lambda: False))


    self.logits = tf.layers.dense(inputs=_activations, units=vocab_size, activation=None)


################################################################################


class Optimizer(object):
  """
  the root class for an optimizer, which is responsible for taking a model
  and training or evaluating it using given optimization procedures (SGD, ADAM, etc.)
  and based on some defined loss metric.
  """
  def __init__(self, inputs, predictions, targets, lrate, loss_name):
    """
    inputs: an input, usually tensorm in the appropriate shape, optimizer doesn't care about the shape though
    preditions: predictions to be used to compute the loss
    targets: labels in the appropriate format based on the loss
    lrate: learning rate
    loss_name: string name of a loss function, such as softmax_cross_entropy and mse.
    """
    self.lrate  = lrate
    self.inputs = inputs
    self.predictions = predictions
    self.targets = targets
    self.loss_name = loss_name
    self.history = []

    N = tf.shape(predictions)[0]

    if loss_name == 'mse':
      self.loss = tf.reduce_sum(tf.square(predictions - targets))
      self.acc = None
    elif loss_name == 'softmax_cross_entropy':
      self.loss = tf.losses.softmax_cross_entropy(onehot_labels=targets, logits=predictions)
      self.acc = tf.contrib.metrics.accuracy(tf.argmax(targets, 1), tf.argmax(predictions,1))
    else:
      raise Exception("Loss name not recognized: {}".format(loss_name))


  def compile_train_op(self, optimizer_name, vars_to_train, max_grad_norm=None):
    """
    Returns a training op that can be called on a batch to update the weights.
    """
#    grads = tf.gradients(self.loss, vars_to_train)
#
#    if max_grad_norm is not None:
#      grads, _ = tf.clip_by_global_norm(grads, max_grad_norm)

    if optimizer_name=='adam':
        optimizer = tf.train.AdamOptimizer(self.lrate)
    elif optimizer_name=='rmsprop':
        optimizer = tf.train.RMSPropOptimizer(self.lrate)
    elif optimizer_name=='sgd':
        optimizer = tf.train.GradientDescentOptimizer(self.lrate)
    elif optimizer_name=='momentum':
        optimizer = tf.train.MomentumOptimizer(self.lrate, momentum=0.01)
    else:
        raise Exception("Optimizer {} not supported.".format(optimizer_name))

    #self.train_op = optimizer.apply_gradients(zip(grads, vars_to_train))
    self.train_op = optimizer.minimize(self.loss)
    return self.train_op


  def go_over_batch(self, x, y, tf_session, train=False, feed_dict={}):
    """
    Goes over one batch of data given as x and y, and returns the loss. If 'train'
    is specified as True, it will update the weights by calling the train op.
    The order that metrics and train op are run is unknown because a dict is used.
    Takes:
      feed_dict: can be passed from outside to update extra stuff related to the model(s)
    Returns:
      a dictionary containing the 'loss' or 'acc' if applicable
    """
    feed_dict.update({self.inputs: x, self.targets: y})
    ops_to_run = {'loss': self.loss, 'predictions': self.predictions}

    if self.acc is not None:
      ops_to_run['acc'] = self.acc

    if train:
      ops_to_run['train_op'] = self.train_op

    return tf_session.run(ops_to_run, feed_dict=feed_dict)


  def go_over_data(self, dataX, dataY, batch_size, tf_session, train=False,
                   feed_dict={}, save_in_hist=True, full_pred=False):
    """
    Calls go_over_batch over the given data set batched by given batch size.
    Takes:
      feed_dict: can be passed from outside to update extra stuff related to the model(s)
      save_in_hist: if true the results will be saved in the history object
      full_pred: if True, a prediction on all examples will be computed after going over each batch
    Returns:
      a history of loss and acc (if applicable) and the means of these values
      for and over all batches. The full history is still saved in self.history.
    """
    result_dict = {'loss_hist': [], 'acc_hist': [], 'pred_hist': []}
    loss_so_far = 0
    acc_so_far = 0

    for x_btch, y_btch in batch_generator((dataX, dataY), batch_size, allow_smaller_last_batch=True, shuffle=True):
      _batch_result = self.go_over_batch(x_btch, y_btch, tf_session, train=train, feed_dict=feed_dict)

      if full_pred:
        _all_preds = self.go_over_data(dataX, dataY, dataY.shape[0], tf_session, False, feed_dict, False, False)['pred_hist']
        result_dict['pred_hist'].append([i for x in _all_preds for i in x])
      else:
        result_dict['pred_hist'].append(_batch_result['predictions'])

      result_dict['loss_hist'].append(_batch_result['loss'])
      loss_so_far += _batch_result['loss']

      if 'acc' in _batch_result:
        result_dict['acc_hist'].append(_batch_result['acc'])
        acc_so_far += _batch_result['acc']

    result_dict['mean_loss'] = loss_so_far/dataY.shape[0]
    result_dict['mean_acc'] = acc_so_far/dataY.shape[0]

    if save_in_hist:
      self.history.append(result_dict)

    return result_dict


  def fit(self):
    pass
    """
    Similar to Keras' fit, it will handle batching and loss accumulation all in one.
    """


