### Module containing functions needed to process patient data and prepare
### them for downstream machines learning purposes.

### Sadegh M, Summer-Fall 2016-17

import datetime as dt
import numpy as np
import csv
import pickle
import xml.etree.cElementTree as XMLParser


def add_retrsopect_feature(dataRaw, lookback, bgl_ind=0):
  """
  adds a feature that is equal to the bgl value that is 'lookback' time units
  before t0. Modifies the dataRaw input. The first dimension in dataRaw[pID] is
  assumed to be N, the number of examples.
  Takes:
    bgl_ind: the index of bgl in dataRaw
  """
  for pID in dataRaw:
    N = dataRaw[pID].shape[0]
    dataRaw[pID] = np.concatenate((dataRaw[pID], dataRaw[pID][range(-lookback,N-lookback), bgl_ind].reshape(-1,1)), axis=1)

  return dataRaw


def add_random_feature(dataRaw):
  """
  adds a uninormly distributed random feature with values between 0 and 1.
  """
  for pID in dataRaw:
    N = dataRaw[pID].shape[0]
    dataRaw[pID] = np.concatenate((dataRaw[pID], np.random.rand(N, 1)), axis=1)

  return dataRaw


def add_zero_feature(dataRaw):
  """
  adds a feature with all values equal to zero.
  """
  for pID in dataRaw:
    N = dataRaw[pID].shape[0]
    dataRaw[pID] = np.concatenate((dataRaw[pID], np.zeros((N, 1))), axis=1)

  return dataRaw


def add_extra_feature(dataRaw, cats_to_load):
  if cats_to_load:
    for catlabel in cats_to_load:
        if catlabel.startswith('retrospect'):
          lookback = int(catlabel.partition('_')[2])
          print("\n > Including a retrospect feature for all patients with lookback", lookback)
          dataRaw = add_retrsopect_feature(dataRaw, lookback, 0)
        elif catlabel == 'random':
          print('\n > Including a random feature for all patients...')
          dataRaw = add_random_feature(dataRaw)
        elif catlabel == 'zero':
          print('\n > Including a zero feature for all patients...')
          dataRaw = add_zero_feature(dataRaw)
  return dataRaw


def load_synthetic(fileNamesAndIDs, res=5):
  """ loads synthetic data from given files. Being synthetic, these data do not
      belong to a patient, but the output format is similar to the load_from_txt function
      in order to be compatible with other downstream data processing functions """
  data = {}
  times = {}
  t = 0

  for ID, fname in fileNamesAndIDs:
      data[ID] = []
      times[ID] = []
      f = open(fname, 'r')
      print("\n> Parsing file {} for set '{}'...".format(fname, ID))
      L = f.readline()
      while L:
          data[ID].append([float(i) for i in L.split()])
          times[ID].append(t)
          t += res
          L = f.readline()
      data[ID] = np.asarray(data[ID])

  return data, times


def load_from_csv(fileNamesAndIDs, delimiter=',', pickOneOfEach=1,
                  takeAverage=True, excludeBGLFromAveraging=True,
                  includeCats=['bgl'],
                  dtFormat='%Y-%m-%d %H:%M:%S'):
  """Reads data from csv files. The CSV header is used as category names, except
  for timeOfDay which need not be in the header with this name.
  TAKES:
    fileNamesAndIDs: a list of pairs of (patiend ID, filename) to process
    pickOneOfEach: will pick only one of this many data points, determines resolution
    takeAverage: if True, uses the average of pickOneOfEach elements as representative
    excludeBGLFromAveraging: if True, BGL is excluded from averaging
    includeCats: read only these categories from the files, this is a list
  RETURNS:
  """
  if not isinstance(includeCats, list):
    raise Exception("'includeCats' must be a LIST of category labels.")

  data = {}
  timeStamps = {}
  isRealBGL = {}
  ibgl = includeCats.index('bgl')

  for pID, fname in fileNamesAndIDs:
    print("\n> Parsing file {} for patient '{}'...".format(fname, pID))
    data[pID] = []
    timeStamps[pID] = []
    isRealBGL[pID] = []

    with open(fname, 'r') as csvfile:
      csvrows = list(csv.reader(csvfile, delimiter=delimiter))

    header = csvrows[0]
    csvrows = csvrows[1:]
    inds = dict(zip(header, range(len(header))))
    i_begin = 0
    numRows = len(csvrows)
    while i_begin <= numRows - pickOneOfEach:
      i_end = i_begin + pickOneOfEach - 1
      curRows = np.asarray([[float(row[inds[catlabel]])
                            for catlabel in includeCats
                            if not catlabel.startswith('timeOf')]
                      for row in csvrows[i_begin:i_end+1]])
      if takeAverage:
        data[pID].append(np.sum(curRows, axis=0)/curRows.shape[0])
        if excludeBGLFromAveraging is True:
          data[pID][-1][ibgl] = curRows[-1, ibgl]
      else:
        data[pID].append(curRows[0, :])
      timeStamps[pID].append(csvrows[i_end][inds['date']]+' '+csvrows[i_end][inds['time']])
      isRealBGL[pID].append(True)
      i_begin += pickOneOfEach

    data[pID] = np.asarray(data[pID])

    intvls = {'timeOfDay':24*3600, 'timeOfWeek':7*24*3600}
    timeOfX = None

    for catlabel in includeCats:
      if catlabel.startswith('timeOf'):
        if catlabel.endswith('_real'):
          print(" > Category {} is being included as total seconds since reference...".format(catlabel))
          timeOfX = []
          interval_secs = intvls[catlabel.partition('_')[0]]
          t1 = dt.datetime.strptime('01-01 00:00:00', '%m-%d %H:%M:%S')
          for i, ts in enumerate(timeStamps[pID]):
            t2 = dt.datetime.strptime(ts.partition('-')[2], '%m-%d %H:%M:%S').replace(second=0)
            timeOfX.append([(t2 - t1).total_seconds() % interval_secs])

        if 'binary' in catlabel:
          nFeats = int(catlabel.split('_')[-1])
          print(" > Time of day is being included as a {}-way binary features...".format(nFeats))
          timeOfX = []
          interval_secs = intvls[catlabel.partition('_')[0]]
          t1 = dt.datetime.strptime('01-01 00:00:00', '%m-%d %H:%M:%S')
          for i, ts in enumerate(timeStamps[pID]):
            t2 = dt.datetime.strptime(ts.partition('-')[2], '%m-%d %H:%M:%S').replace(second=0)
            _thisfeat = [0]*nFeats
            _thisfeat[int(nFeats*((t2 - t1).total_seconds()%interval_secs)/interval_secs)] = 1
            timeOfX.append(_thisfeat)

        if catlabel.endswith('_ticks'):
          print(" > Time of day is being included as one feature with ticks on the hour...")
          timeOfX = []
          t1 = '?'
          for ts in timeStamps[pID]:
            hrStr = ts.split(' ')[1].split(':')[0]
            t2 = hrStr
            timeOfX.append([(t2 != t1)*1])
            t1 = hrStr

        data[pID] = np.concatenate((data[pID], timeOfX), axis=1)

  return data, timeStamps, isRealBGL



def load_from_xml(fileNamesAndIDs, res=5, includeCats=[],
                  dtFormat='%d-%m-%Y %H:%M:%S', verbose=True):
  """ Loads patient data from an XML file containing all the bgl and even information.
  The first category in the file must be BGL values (anything before bgl will be skipped)
  because other catogires might depend on them. The other categories can be in any order.
  NOTE: hypo action should occur after meals in the input file, also temp basal after basal
  NOTE: when 'meal' is included in the categories, hypo_actions are considered meals
  TAKES:
	  fileNamesAndIDs: a list of pairs of (patiend ID, filename) to process
	  res: resolution in minutes of the output sequences. Default is 5 minutes for
		  blood glucose measurements.
	  includeCats: include these categories. It must be a list of categories.
		  Note that 'bgl' is always automatically added.
	  dtFormat: expected date time format in the XML file.
  RETURNS:
	  data: a dictionary of {pID:X} where X is an N x D list, N: number of bgl data and
	  timeStamps: the time stamps corresponding to BGL data
	  isRealBGL: a dict of binary lists indicating whether each sample has a real BGL or not (an interpolated one)
  """
  data = {}
  timeStamps = {}
  isRealBGL = {}
  
  def _valueAtTimeInCat(tm, category, threshold):
    """
    Helper function that searches in a 'category' (which is just a dictionary of
    datetime:value pairs) and checks if 'tm' is in the keys within a 'threshold'.
    """
    delta1 = dt.timedelta(minutes=1)
    for r in range(threshold+1):
        if (tm + r*delta1) in category:
            return category[tm + r*delta1]
        elif (tm - r*delta1) in category:
            return category[tm - r*delta1]
    return 0

  if not isinstance(includeCats, list):
	  raise Exception('The category list to include must be a LIST of category labels.')


  for pID, fname in fileNamesAndIDs:
	  print("\n> Parsing file {} for patient '{}'...".format(fname, pID))
	  xmlRoot = XMLParser.parse(fname).getroot()
	  xmlIter = xmlRoot.iter()

	  item = next(xmlIter)
	  while item.tag != 'glucose_level':
		  item = next(xmlIter)

	  print(" > Parsing category 'bgl'...")
	  categories      = [] # list of encountered categories
	  bgl             = []
	  isRealBGL[pID]  = []
	  categories.append(bgl)
	  timeStamps[pID] = []

	  item = next(xmlIter)
	  prevLevel = 0
	  prevDT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)

	  while item.tag == 'event':
		  curDT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
		  curLvl = float(item.attrib['value'])

		  diff = (curDT - prevDT).total_seconds() / 60

		  if diff > res:
			  if verbose:
				  print(" Warning: bgl data discontinuity at {}".format(prevDT))
			  s = int(diff // res - (diff % res == 0))
			  m = (curLvl - prevLevel) / diff
			  for i in range(s):
				  prevLevel += m * res
				  bgl.append(prevLevel)
				  isRealBGL[pID].append(False)
				  timeStamps[pID].append(prevDT + dt.timedelta(minutes=res*(i+1)))

		  bgl.append(curLvl)
		  isRealBGL[pID].append(True)
		  timeStamps[pID].append(curDT)

		  prevDT = curDT
		  prevLevel = curLvl
		  item = next(xmlIter)

	  if 'timeOfDay_real' in includeCats:
		  print(" > Time of day is being included as total seconds since midnight...")
		  timeOfDay_real = {ts:(ts-ts.replace(minute=0, hour=0)).total_seconds() for ts in timeStamps[pID]}
		  categories.append(timeOfDay_real)

	  if 'timeOfDay_ticks' in includeCats:
		  print(" > Time of day is being included as one feature with ticks on the hour...")
		  timeOfDay_ticks = {}
		  prevHour = -1
		  for ts in timeStamps[pID]:
			  thisHour = ts.hour
			  timeOfDay_ticks[ts] = (thisHour != prevHour)*1
			  prevHour = ts.hour

		  categories.append(timeOfDay_ticks)

	  while item.tag:
		  try:
			  if item.tag=='finger_stick' and 'finger_stick' in includeCats:
				  print(" > Parsing category 'finger_stick'...")
				  fstick = {}
				  categories.append(fstick)
				  item = next(xmlIter)
				  while item.tag=='event':
					  DT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
					  fstick[DT] = float(item.attrib['value'])
					  item = next(xmlIter)

			  elif item.tag=='basal' and 'basal' in includeCats:
				  print(" > Parsing category 'basal'...")
				  basal = {}
				  categories.append(basal)
				  beginDT = None
				  item = next(xmlIter)
				  _value = None
				  while item.tag=='event':
					  endDT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
					  if beginDT:
						  while beginDT <= endDT:
							  basal[beginDT] = _value
							  beginDT = beginDT + dt.timedelta(minutes=1)
					  beginDT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
					  _value = item.attrib['value']
					  item = next(xmlIter)

			  elif item.tag=='temp_basal' and 'basal' in includeCats:
				  print(" > Parsing category 'temp_basal' and correcting 'basal'...")
				  item = next(xmlIter)
				  while item.tag=='event':
					  beginDT = dt.datetime.strptime(item.attrib['ts_begin'], dtFormat).replace(second=0)
					  endDT = dt.datetime.strptime(item.attrib['ts_end'], dtFormat).replace(second=0)
					  while beginDT <= endDT:
						  basal[beginDT] = item.attrib['value']
						  beginDT = beginDT + dt.timedelta(minutes=1)
					  item = next(xmlIter)

			  elif item.tag=='bolus' and 'bolus' in includeCats:
				  print(" > Parsing category 'bolus'...")
				  bolus = {}
				  categories.append(bolus)
				  item = next(xmlIter)
				  while item.tag=='event':
					  beginDT = dt.datetime.strptime(item.attrib['ts_begin'], dtFormat).replace(second=0)
					  endDT = dt.datetime.strptime(item.attrib['ts_end'], dtFormat).replace(second=0)
					  num_minutes = max(1, (endDT-beginDT).total_seconds() / 60)
					  bolus_per_minute = float(item.attrib['dose']) / num_minutes
					  while beginDT <= endDT:
						  bolus[beginDT] = min(res, num_minutes) * bolus_per_minute
						  beginDT = beginDT + dt.timedelta(minutes=res)
						  num_minutes = (endDT-beginDT).total_seconds() / 60
					  item = next(xmlIter)

			  elif item.tag=='meal' and 'meal' in includeCats:
				  print(" > Parsing category 'meal'...")
				  meal = {}
				  categories.append(meal)
				  item = next(xmlIter)
				  while item.tag=='event' or item.tag=='food':
					  if item.tag=='event':
						  DT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
						  carb = float(item.attrib['carbs'])
						  meal[DT] = carb
					  item = next(xmlIter)

			  elif item.tag=='sleep' and 'sleep' in includeCats:
				  print(" > Parsing category 'sleep'...")
				  sleep = {}
				  categories.append(sleep)
				  item = next(xmlIter)
				  while item.tag=='event':
					  beginDT = dt.datetime.strptime(item.attrib['ts_end'], dtFormat).replace(second=0)
					  endDT = dt.datetime.strptime(item.attrib['ts_begin'], dtFormat).replace(second=0)
					  if beginDT > endDT:
						  beginDT, endDT = endDT, beginDT
					  if endDT-beginDT > dt.timedelta(hours=12) and verbose:
						  print(" Warning: long sleep detected from {} to {}, duration={} hr"
								  .format(beginDT, endDT, (endDT-beginDT).total_seconds()/3600))
					  while beginDT <= endDT:
						  sleep[beginDT] = 1
						  beginDT = beginDT + dt.timedelta(minutes=1)
					  item = next(xmlIter)

			  elif item.tag=='work' and 'work' in includeCats:
				  print(" > Parsing category 'work'...")
				  work = {}
				  categories.append(work)
				  item = next(xmlIter)
				  while item.tag=='event':
					  beginDT = dt.datetime.strptime(item.attrib['ts_begin'], dtFormat).replace(second=0)
					  endDT = dt.datetime.strptime(item.attrib['ts_end'], dtFormat).replace(second=0)
					  if endDT-beginDT > dt.timedelta(hours=12) and verbose:
						  print(" Warning: long work detected from {} to {}, duration={} hr"
								  .format(beginDT, endDT, (endDT-beginDT).total_seconds()/3600))
					  while beginDT <= endDT:
						  work[beginDT] = float(item.attrib['intensity'])
						  beginDT = beginDT + dt.timedelta(minutes=1)
					  item = next(xmlIter)

			  elif item.tag=='infusion_set' and 'infusion_set' in includeCats:
				  print(" > Parsing category 'infusion_set'...")
				  infusion_set = {}
				  categories.append(infusion_set)
				  item = next(xmlIter)
				  while item.tag=='event':
					  DT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
					  infusion_set[DT] = 1
					  item = next(xmlIter)

			  elif item.tag=='hypo_event' and 'hypo_event' in includeCats:
				  print(" > Parsing category 'hypo_event'...")
				  hypo_event = {}
				  categories.append(hypo_event)
				  item = next(xmlIter)
				  while item.tag=='event' or item.tag=='symptom':
					  if item.tag=='event':
						  DT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
						  hypo_event[DT] = 1
					  item = next(xmlIter)

			  elif item.tag=='hypo_action' and 'meal' in includeCats:
				  print(" > Parsing category hypo_action and adding to 'meal...")
				  item = next(xmlIter)
				  while item.tag=='event':
					  DT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
					  if DT in meal and verbose:
						  print(" Warning: hypo action already in 'meals' at {} with carbs={}, replacing with {}..."
								  .format(DT, meal[DT], item.attrib['carbs']))
					  meal[DT] = item.attrib['carbs']
					  item = next(xmlIter)

			  elif item.tag=='exercise' and 'exercise' in includeCats:
				  print(" > Parsing category 'exercise'...")
				  exercise = {}
				  categories.append(exercise)
				  item = next(xmlIter)
				  while item.tag=='event':
					  DT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
					  duration = float(item.attrib['duration'])
					  intensity = float(item.attrib['intensity'])
					  for i in range(int(duration)+1):
						  exercise[DT+dt.timedelta(minutes=i)] = intensity
					  item = next(xmlIter)

			  elif item.tag=='basis_heart_rate' and 'basis_heart_rate' in includeCats:
				  print(" > Parsing category 'basis_heart_rate'...")
				  hrate = {}
				  categories.append(hrate)
				  item = next(xmlIter)
				  while item.tag=='event':
					  DT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
					  hrate[DT] = float(item.attrib['value'])
					  item = next(xmlIter)

			  elif item.tag=='basis_gsr' and 'basis_gsr' in includeCats:
				  print(" > Parsing category 'basis_gsr'...")
				  gsr = {}
				  categories.append(gsr)
				  item = next(xmlIter)
				  while item.tag=='event':
					  DT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
					  gsr[DT] = float(item.attrib['value'])
					  item = next(xmlIter)

			  elif item.tag=='basis_skin_temperature' and 'basis_skin_temperature' in includeCats:
				  print(" > Parsing category 'basis_skin_temperature'...")
				  skin_temp = {}
				  categories.append(skin_temp)
				  item = next(xmlIter)
				  while item.tag=='event':
					  DT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
					  skin_temp[DT] = float(item.attrib['value'])
					  item = next(xmlIter)

			  elif item.tag=='basis_air_temperature' and 'basis_air_temperature' in includeCats:
				  print(" > Parsing category 'basis_air_temperature'...")
				  air_temp = {}
				  categories.append(air_temp)
				  item = next(xmlIter)
				  while item.tag=='event':
					  DT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
					  air_temp[DT] = float(item.attrib['value'])
					  item = next(xmlIter)

			  elif item.tag=='basis_steps' and 'basis_steps' in includeCats:
				  print(" > Parsing category 'basis_steps'...")
				  steps = {}
				  categories.append(steps)
				  item = next(xmlIter)
				  while item.tag=='event':
					  DT = dt.datetime.strptime(item.attrib['ts'], dtFormat).replace(second=0)
					  steps[DT] = float(item.attrib['value'])
					  item = next(xmlIter)

			  else:
				  item = next(xmlIter)

		  except StopIteration:
			  break

	  print(" > Parsing done, now merging the data into a single matrix...")

	  N = len(bgl)
	  D = len(categories)
	  data[pID] = np.zeros((N, D))

	  for t, tm in enumerate(timeStamps[pID]):
		  data[pID][t, 0] = bgl[t]
		  for c, cat in enumerate(categories[1:]):
			  data[pID][t, c+1] = _valueAtTimeInCat(tm, cat, threshold=2)

  return data, timeStamps, isRealBGL


def load_from_KalmanPatientXML(fileNamesAndIDs, times, includeCats=[]):
  """
  Loads the patient data from an XML file created by the KalmanPatient method.
  The file is assumed to contain BGL, meal, basal and bolus data with one-minute
  resolution.
  TAKES:
      fileNamesAndIDs: an iterator containing pairs of (patient ID, filename) to process
      times: time stamps corresponding to bgl values, data will be loaded at these times
      includeCats: similar to 'load_from_txt'
  RETURNS:
      Similar to the function 'load_from_txt'
  Note: Of all the meals and boluses in a 5 minute interval, a maximum of two
  can be considered.
  """
  data = {}
  dtFormat = '%d-%m-%Y %H:%M:%S'

  if not isinstance(includeCats, list):
      raise Exception('The category list to include must be a LIST of category labels.')

  D = len(includeCats)

  for pID, fname in fileNamesAndIDs:
      print("\n> Parsing file {} for patient '{}'...".format(fname, pID))
      xmlRoot = XMLParser.parse(fname).getroot()
      curData = []

      xmlIter = xmlRoot.iter()
      prevDT, DT = None, None
      for ts in times[pID]:
          thisDelta = [0]*D
          while prevDT != ts:
              try:
                  item = next(xmlIter)
              except StopIteration:
                  prevDT = DT
              if item.tag == 'delta':
                  prevDT = DT
                  DT = dt.datetime.strptime(item.attrib['end'], dtFormat)
              elif item.tag != 'patient':
                  if item.tag in includeCats:
                      _idx = includeCats.index(item.tag)
                      if item.tag in ['meal', 'bolus']:
                          if ts-DT > dt.timedelta(minutes=2) and curData[-1][_idx]==0:
                              curData[-1][_idx] = float(item.text)
                          else:
                              thisDelta[_idx] = float(item.text)
                      else:
                          thisDelta[_idx] = float(item.text)
          curData.append(thisDelta)


      data[pID] = np.asarray(curData)

  return data



def prepare_for_learning(dataRaw, timeStampsRaw, realBGLMask=None, otherMask=None, seqLen=1,
          btch_sz=1, predHorizon=1, dType='float32', bgl_scale=1.0, other_scale=1.0,
          testPortion=0.2, validPortion=0.0, normalize=True, stateful=0,
          dTransp=[0,1,2], lTransp=[0,1], useLabelDiff=False, filterWhatIfPoints=False,
          filterPreBreak=False, whatIfFeatures=(1,2), whatIfScenario=False):
  """
      Prepares a given data set for stateful or stateless RNN learning.
      TAKES:
          dataRaw: a dictionary where dataRaw[pID] is a NxD matrix of raw data for patient pID
          timeStampsRaw: the time stamps associated with entries in dataRaw
          realBGLMask: specifies which BGL values are real as opposed to interpolated.
          otherMask: an extra mask that is applied when choosing final sequence examples
          dTransp: order by which default 'data' tensors will be transposed
          lTransp: order by which default 'label' tensors will be transposed
          bgl_scale and normalize: bgl_scale is used for bgl values independently from normalize
           which makes all other values between 0 and 1
          other_scale: scales other values whether normalized or not
          validPortion, testPortion: if less than 1.0 are taken as percentages, if absolute numbers,
            are taken as number of points. If dicts, are taken one value per pID.
            Test portion is taken from the end of the data. Then valid portion is taken from the remaining.
          useLabelDiff: if True, the difference of t0+horiz and t0 values is saved as label instead of t0+horiz
          filterWhatIfPoints: a bool, for whether points with what-if events (i.e.
            meal and bolus events happening between t0 and t0+pred_horiz) should be filtered.
          filterPreBreak: If True applies realBGLMask and WhatIf filters before
            breaking into train, valid and test. False filters after and this
            may result in smaller valid and test portions than specified.
          whatIfFeatures: indices of the events in the data that are considered
            for filtering what-if events
          whatIfScenario: if True, allows sequences to be from t0-seq_len to t0+pred_horiz instead of to t0
      RETURNS:
          data: a dictionary where data[pID]['train'] is a (N x SEQ_LEN x D) matrix for patient pID.
              'data' also contains 'timeStamps' and 'indices' of each data point
  """

  data = {}
  incMask_dict = {}

  if (whatIfScenario and filterWhatIfPoints):
      raise Warning("\nWarning: whatif scenario is enabled but whatif points are being filtered!\n")

  if whatIfScenario:
      seqLenTail = predHorizon
  else:
      seqLenTail = 0

  for pID in dataRaw.keys():
      dataRaw[pID][:, 0] *= bgl_scale

      if normalize:
          b = dataRaw[pID].min(axis=0)
          s = dataRaw[pID].max(axis=0) - b
          s[s==0.] = 1 # to prevent nan values
          b[0], s[0] = 0, 1 # do not normalize bgl data
          dataRaw[pID] = (dataRaw[pID] - b) / s

      dataRaw[pID][:, 1:] *= other_scale

      _m = max(0, predHorizon)
      N = dataRaw[pID].shape[0] - _m - seqLen + 1
      D = dataRaw[pID].shape[1]

      print("\n> Processing {} data points with Dim {}, seq_len {} and predHorizon {} for patient '{}'".format(N, D, seqLen, predHorizon, pID))

      if filterWhatIfPoints is True:
        print("  Note: What-if points are being filtered...")

      if otherMask is not None:
        print("  Note: Custom mask is provided...")
        assert(len(otherMask[pID]) == N)

      allData = np.zeros((N, seqLen+seqLenTail, D), dtype=dType)
      allLabels = np.zeros((N, 1), dtype=dType)
      allTimeStamps = np.zeros(N, dtype='<U19')
      inclusionMask = np.ones(N, dtype=bool)

      ki = 0  # dummy var used to track interpolated BGL indices if such mask is provided

      for i in range(N):
          t0 = i + seqLen - 1

          allData[i, :, :] = dataRaw[pID][i:t0+seqLenTail+1, :]

          if useLabelDiff:
              allLabels[i, :] = [dataRaw[pID][t0+predHorizon, 0] - dataRaw[pID][t0, 0]]
          else:
              allLabels[i, :] = [dataRaw[pID][t0+predHorizon, 0]]

          allTimeStamps[i] = timeStampsRaw[pID][t0]

          if filterWhatIfPoints is True and np.any(dataRaw[pID][t0+1:t0+predHorizon+1, whatIfFeatures]):
              inclusionMask[i] = False

          if realBGLMask and pID in realBGLMask:
              inclusionMask[i] = inclusionMask[i] and realBGLMask[pID][t0+predHorizon]
              if not realBGLMask[pID][t0]:
                  ki += 1
                  slp = dataRaw[pID][t0-ki, 0] - dataRaw[pID][t0-ki-1, 0]
                  extrapol = dataRaw[pID][t0-ki, 0] + slp * (np.arange(ki+seqLenTail) + 1)
                  _max_ki_seqlen = min(ki, seqLen)+seqLenTail
                  allData[i, -_max_ki_seqlen:, 0] = extrapol[-_max_ki_seqlen:]
              else:
                  ki = 0

          if otherMask and pID in otherMask:
              inclusionMask[i] = inclusionMask[i] and otherMask[pID][i]

      ### NOTE: stateful code might be stale
      # if stateful:
      #     N -= N % btch_sz  # to make it divisible by batch size (stateful requirement)
      # if (stateful == 1 and seqLen != btch_sz and seqLen != 1):
      #     raise Exception("sequence length and batch size should be the same with Stateful=1, unless sequence length is 1")
      #
      # if (stateful == 1 and seqLen != btch_sz and seqLen == 1) or (stateful == 2):
      #     # rearranging the data so that they satisfy stateful requirements
      #     print('> Rearranging the data to satisfy stateful requirements...')
      #     allData = allData.reshape(btch_sz, -1, seqLen, D)
      #     allData = allData.transpose(1, 0, 2, 3)
      #     allData = allData.reshape(-1, allData.shape[2], allData.shape[3])
      #
      #     allLabels = allLabels.reshape(btch_sz, -1, allLabels.shape[1])
      #     allLabels = allLabels.transpose(1, 0, 2)
      #     allLabels = allLabels.reshape(-1, allLabels.shape[2])


      if filterPreBreak:
          allData = allData[inclusionMask,:,:]
          allLabels = allLabels[inclusionMask,:]
          N = allData.shape[0]

      if isinstance(testPortion, dict):
        testPortionPID = testPortion[pID]
      else:
        testPortionPID = testPortion

      if isinstance(validPortion, dict):
        validPortionPID = validPortion[pID]
      else:
        validPortionPID = validPortion

      if testPortionPID < 1.0:
          testPortionPID = testPortionPID*N

      if validPortionPID < 1.0:
          validPortionPID = validPortionPID*N

      idx_te = N - int(testPortionPID) + _m
      idx_va = idx_te - int(validPortionPID)
      slice_tr = slice(None, idx_va-_m)
      slice_va = slice(idx_va+seqLen-1, idx_te-_m)
      slice_te = slice(idx_te+seqLen-1, None)

      if filterPreBreak:
          incMask_all = np.ones(N, dtype=bool)
      else:
          incMask_all = inclusionMask

      incMask_tr = incMask_all[slice_tr]
      incMask_va = incMask_all[slice_va]
      incMask_te = incMask_all[slice_te]


      data[pID] = {}
      data[pID]['all']      = allData[incMask_all, :, :].transpose(dTransp)
      data[pID]['train']    = allData[slice_tr][incMask_tr, :, :].transpose(dTransp)
      data[pID]['validate'] = allData[slice_va][incMask_va, :, :].transpose(dTransp)
      data[pID]['test']     = allData[slice_te][incMask_te, :, :].transpose(dTransp)

      data[pID]['allLabels']      = allLabels[incMask_all, :].transpose(lTransp)
      data[pID]['trainLabels']    = allLabels[slice_tr][incMask_tr, :].transpose(lTransp)
      data[pID]['validateLabels'] = allLabels[slice_va][incMask_va, :].transpose(lTransp)
      data[pID]['testLabels']     = allLabels[slice_te][incMask_te, :].transpose(lTransp)

      data[pID]['timeStamps'] = allTimeStamps[inclusionMask]

      incMask_dict[pID] = inclusionMask

      N_tr = data[pID]['train'].shape[0]
      N_va = data[pID]['validate'].shape[0]
      N_te = data[pID]['test'].shape[0]
      print("  N_train: {}  N_validation: {}  N_test: {}  Total: {}".format(N_tr, N_va, N_te, N_tr+N_va+N_te))

      _fullrange = np.arange(N_tr + N_va + N_te)
      data[pID]['allIndices'] = _fullrange
      data[pID]['trainIndices'] = _fullrange[:N_tr]
      data[pID]['validateIndices'] = _fullrange[N_tr:N_tr+N_va]
      data[pID]['testIndices'] = _fullrange[N_tr+N_va:]

#  save the inclusion masks for later use
#  with open('whatIfMask_60min.pkl', 'wb') as f:
#        pickle.dump(incMask_dict, f)

  return data, incMask_dict



def get_200_indices(data, dtFileNames, dtFormat = '%Y-%m-%d %H:%M:%S'):
  """
  ??
  Note: does a linear search for each item so order in .ts files does not matter.
  Note: Negative indices mean that the given time stamp is not in the prepared data.
  """

  indices = {}
  delta1 = dt.timedelta(minutes=1)
  for pID, fName in dtFileNames:
      with open(fName, 'r') as f:
          ftimes = [dt.datetime.strptime(L.strip(), dtFormat) for L in f.readlines()]
      indices[pID] = []
      category = data[pID]['timeStamps']
      for ft in ftimes:
          for r in range(3):
              w = np.where(category==str(ft + r*delta1))[0]
              if len(w) > 0:
                  indices[pID].append(w[0])
              else:
                  w = np.where(category==str(ft - r*delta1))[0]
                  if len(w) > 0:
                    indices[pID].append(w[0])
  return indices



def mix_data(data, profile, IDoffset=None, AxisOfN=0, shuffle=False, randgen=None):
  """
  mixes the training, test and validation data and labels of the given patients
  and returns as one with or without shuffling.
  NOTE: makes a copy of all the data so better have enough memory!
  TAKES:
      profile: a training profile delineating slices to mix for training, validation and test data.
      data: a dictionary in the form pID:its data. See function load_from_txt.
      IDoffset: the offset to be added to the index of each data point to get its ID
          If None is used the offsets will all be zero (making IDs unusable)
      AxisOfN: the axis in data[pID] corresponding to examples (concatenation happens along this axis).
          This axis is assumed to be 0 in the labels.
      shuffle: if True, shuffles all three parts of the mixed data before returning
      randgen: a random number generator that has a permutation function. if None. numpy.random
          will be used.
  RETURNS:
      outData: a data dictionary similar to 'data', but without the 'all' parameter.
          It also will not have a part (e.g. 'validate') if the part is empty or None in 'profile'.
  """
  outData = {}
  elems = ('train', 'validate', 'test') # these elements are mixed
  slices = [slice(None)]*len(elems)

  if randgen is None:
      randgen = np.random

  if not IDoffset:
      IDoffset = {}
      for pID in data.keys():
          IDoffset[pID] = 0

  for e in elems:
      _eInProfile = profile[e.upper() + '_ON']
      if _eInProfile and isinstance(_eInProfile, dict):
          toConcat = []
          toConcatLs = []
          toConcatIDs = []
          for pID, part in _eInProfile.items():
              if isinstance(part, tuple):
                  slices[AxisOfN] = slice(part[0], part[1])
                  toConcat.append( data[pID]['all'][slices] )
                  toConcatLs.append( data[pID]['allLabels'][part[0]:part[1],:] )
                  toConcatIDs.append( data[pID]['allIndices'][part[0]:part[1]]+IDoffset[pID] )
              elif isinstance(part, str):
                  toConcat.append( data[pID][part] )
                  toConcatLs.append( data[pID][part+'Labels'] )
                  toConcatIDs.append( data[pID][part+'Indices']+IDoffset[pID] )
              else:
                  raise Exception("Invalid 'part' specified in profile: ", part)
          outData[e] = np.concatenate(toConcat, axis=AxisOfN)
          outData[e + 'Labels'] = np.concatenate(toConcatLs, axis=0)
          outData[e + 'IDs'] = np.concatenate(toConcatIDs, axis=0)
          if shuffle:
              N = outData[e].shape[AxisOfN]
              _shuffOrder = randgen.permutation(N)
              slices[AxisOfN] = _shuffOrder
              outData[e] = outData[e][slices]
              outData[e+'Labels'] = outData[e+'Labels'][_shuffOrder, :]
              outData[e+'IDs'] = outData[e+'IDs'][_shuffOrder]

  return outData



def concatData(pIDlist, dataDictList, axis):
  """
  Takes a list of patients in 'pIDlist' and searches a list of data sources
  for data regarding those patients. It then concatenates data from these
  sources (in the given order) along given axis and into one big data dictionary.
  """
  allData = {}
  for pID in pIDlist:
      valdSources = [d[pID] for d in dataDictList if pID in d]

      if valdSources:
          allData[pID] = np.concatenate(valdSources, axis)
      else:
          allData[pID] = []
          raise Warning('Patient {} exists in pIDlist but in none of the given data dicts'.format(pID))

  return allData


def prepadData(dataRaw, realBGLindicator, timeStamps, prepadDict, prepadLength):
  """
  Takes a prepad dictionary (refer to the caller documentation for info) and
  prepads the data with the data specified there. It also prepads realBGLindicator
  and timeStamps accordingly to make them compatible with the data.
  """

  _fiveMins = dt.timedelta(minutes=5)
  for pID in prepadDict:
      _src, _idx = prepadDict[pID]
      prepadDataRaw = np.concatenate((dataRaw[_src][_idx-prepadLength+2:_idx+1,:], dataRaw[pID]), axis=0)
      realBGLindicator[pID] = [True]*(prepadLength-1) + realBGLindicator[pID]
      timeStamps[pID] = list(reversed([timeStamps[pID][0]-_fiveMins*(i+1) for i in range(prepadLength-1)])) + timeStamps[pID]
      dataRaw[pID] = prepadDataRaw

  return dataRaw, timeStamps, realBGLindicator


def batch_generator(data, batch_size=1, allow_smaller_last_batch=True, shuffle=False):
  """
  A generator function that takes several data matries of the same dimension, and yields
  the data batch by batch (as a slice not a copied slice of the original values),
  with slices taken along the first axis.
  TAKES:
    data: a tuple of matrices of same dimensionality, e.g. (dataX, dataY)
  """

  N = data[0].shape[0]

  if shuffle:
    _idx = np.random.permutation(N)
  else:
    _idx = np.arange(N)

  for ibt in range(0, N, batch_size):
    if ibt+batch_size > N:
      if allow_smaller_last_batch:
        idx_slice = _idx[ibt:N]
      else:
        return
    else:
      idx_slice = _idx[ibt:ibt+batch_size]

    yield [data_i[idx_slice] for data_i in data]

def prepare_for_arima(data, out_datafile, out_labelfile, part='test', fmt='%d', concat_patients=False):
  """
  TAKES:
    data: A dictionary that contains data for several patients, broken into train,
        validate and test parts.
    out_datafile and out_labelfile: filenames to write the dataset to (required)
        NOTE that data is appended to these files so if they exist delete them.
    part: The part of data of each patient to write into the files.
    fmt: format for the data to write.
    concat_patients: if True, concatenates the data of all patients into one file, if False
        make a separate file for each patient, tagged with pid
  RETURNS:
    Nothing. But two files are written with the given names. If multivariate, the
    data will be reshaped so that there are as many rows in the text file as there
    are examples in the data, where N is data's first dimension.
  """
  import numpy as np

  if concat_patients:
    openmode = 'ab'
  else:
    openmode = 'wb'

  for pid in data:
    if concat_patients:
      _outX = out_datafile
      _outY = out_labelfile
    else:
      _outX = out_datafile + '_' + str(pid)
      _outY = out_labelfile + '_' + str(pid)

    dfout = open(_outX, openmode)
    lfout = open(_outY, openmode)
    N = data[pid][part].shape[0]

    print("Saving {} examples for patient {}...".format(N, pid))
    np.savetxt(dfout, data[pid][part].reshape(N, -1), fmt=fmt)
    np.savetxt(lfout, data[pid][part+'Labels'][:, 0], fmt=fmt)

    dfout.close()
    lfout.close()


if __name__ == '__main__':
  """ Use the main module here for debugging """






