### an R script to read the data stored in a text file through np.savetxt
### and use ARIMAX modeling to make predictions. The data is assumed to be
### in shape Nx(T*D), where N, T and D are the number, length and dimensionality of sequences.
### If model order is set to "auto", uses auto.arima on each sequence
### to identify the model and saves only the prediction at timestep dt.

library(forecast)

# PARAMETERS
mode            <- "aida"
exovars         <- FALSE    # if TRUE uses exogenous variables
dt              <- 2 		# prediction horizon in minutes / 5
MAX_P           <- 20 		# maximum allowable order for AR when doing auto.arima
modelOrder      <- c(1, 0, 0)	# either "auto" or c(p,d,q)


# MAIN
if (mode == "uva"){
  patient_list    <- c('5_adult#005', '6_adult#006', '7_adult#007', '8_adult#008', '9_adult#009')
  dataFileStem    <- "../Data/ForARIMA/uva/uva_nowhatif_X"
  labelFileStem   <- "../Data/ForARIMA/uva/uva_nowhatif_Y"
  D               <- 3  # number of features to be expected in the file
} else if (mode == "ws500") {
  patient_list    <- c('559', '563', '570', '575', '588', '591')
  dataFileStem    <- "../Data/ForARIMA/ws500/ws500_nointerp_nowhatif_X"
  labelFileStem   <- "../Data/ForARIMA/ws500/ws500_nointerp_nowhatif_Y"
  D               <- 4  # number of features to be expected in the file
} else if (mode == "aida") {
  patient_list    <- c('4', '7', '9', '11', '18', '19', '20', '28', '35', '39')
  dataFileStem    <- "../Data/ForARIMA/aida/aida_raw_X_30min"
  labelFileStem   <- "../Data/ForARIMA/aida/aida_raw_Y_30min"
  D               <- 1  # number of features to be expected in the file
} else if (mode == "300") {
  patient_list    <- c('')
  dataFileStem    <- "../Data/ForARIMA/300/X_nowhatif"
  labelFileStem   <- "../Data/ForARIMA/300/Y_nowhatif"
  D               <- 1  # number of features to be expected in the file
}

rmses <- vector("list", 10)
pind <- 0

for (pid in patient_list){
    pind <- pind + 1
    print(paste("Loading data for patient", pid))
    data <- t(read.table(paste(dataFileStem, "_", pid, sep=""), na.strings=c("","nan","NA")))
    T <- dim(data)[1] / D
    N <- dim(data)[2]
    print(paste(N, "exmaples loaded with length", T, "and dimensionality", D))
    results = NULL

    for (i in 1:N) {
        if (i%%500 == 1) {
          print(paste("Doing sequence ", i, "/", N))
        }

        data_i = data[,i]
        dim(data_i) <- c(D, T)
        data_i <- t(data_i)

        if (exovars) {
          xreg = data_i[,2:D]
          newxreg = matrix(0, nrow=dt, ncol=D-1)
        } else {
          xreg = NULL
          newxreg = NULL
        }

        if (modelOrder == "auto") {
    		model <- auto.arima(data_i[,1], max.p=MAX_P, allowdrift=FALSE, xreg=xreg)
        } else {
    		model <- arima(data_i[,1], order=modelOrder, xreg=xreg)
        }

        pred <- forecast(model, h=dt, newxreg=newxreg)

        results[i] = pred$mean[dt]
    }

    # EVALUATION
    paste(shQuote(results), collapse=", ")
    labels <- read.table(paste(labelFileStem, "_", pid, sep=""), na.strings=c("","nan","NA"))
    rmses[pind] = sqrt(sum((labels[,1]-results)^2) / N)
    print(paste("RMSE is ", rmses[pind]))
    write.table(results, file=paste("preds_", pid, ".txt", sep=""), row.names=FALSE, col.names=FALSE)
}

print(paste("Mean RMSE: ", Reduce("+",rmses)/length(rmses)))
