### This keras script aims at training and evaluating an RNN on the BGL time
### series data loaded from various sources.
### Training is done via profiles. See documentation for utils.get_profiles

### Sadegh M, Summer 2016, 2017

from __future__ import print_function

import numpy as np
import time
import argparse
import pickle
from patient_data import load_from_xml, prepare_for_learning, load_from_csv
from patient_data import mix_data, load_from_KalmanPatientXML, concatData, prepadData
from patient_data import add_extra_feature
from bglBaselines import bgl_time_delay
from keras import backend as krs_backend
from shared_args import parse_cmd_arguments, customize_args, get_profiles
from nn_models_base import KerasSimpleRnn
from nn_models_keras import DoubleLstm


args = parse_cmd_arguments(argparse.ArgumentParser())

if args.random_seed is not None:
    np.random.seed(args.random_seed)

args.seq_len = args.seq_len_in_minutes // args.data_resolution
args.pred_horizon = args.pred_horizon_in_minutes // args.data_resolution
args.cats_to_load = eval(args.cats_to_load)
args.other_cats_to_add = eval(args.other_cats_to_add)

args = customize_args(args)

######## PREPARATION ########
if args.data_pkl != '':
    print('\nLoading data from pickle file {}\n'.format(args.data_pkl))
    with open(args.data_pkl, 'rb') as f:
        data = pickle.load(f)
elif 'data' not in globals():
    print('Loading data for patients ', args.pid_list)

    if args.mode == 'synthetic':
        dataRaw, timeStamps, isRealBGL = load_from_csv(zip(args.pid_list, args.data_files), ' ',
                                                        args.data_resolution, True, True,
                                                        args.cats_to_load)
    else:
        dataRaw, timeStamps, isRealBGL = load_from_xml(zip(args.pid_list, args.data_files), args.data_resolution,
                                                        args.cats_to_load, verbose=False)
    if args.prepad_dic:
        print('\nPrepadding the loaded raw data using given sources...')
        dataRaw, timeStamps, isRealBGL = prepadData(dataRaw, isRealBGL, timeStamps, args.prepad_dic,
                                                    prepadLength=args.seq_len)

    dataRaw = add_extra_feature(dataRaw, args.other_cats_to_add)
	
    if args.data_files_test:
        print("Loading test files separately from ", args.data_files_test)
        dataRaw_test, timeStamps_tes, isRealBGL_tes = load_from_xml(zip(args.pid_list, args.data_files_test), args.data_resolution,
                                                        args.cats_to_load, verbose=False)
        dataRaw = concatData(args.pid_list, (dataRaw, dataRaw_test), axis=0)
        args.test_portion = {}
        for pid in args.pid_list:
            timeStamps[pid] += timeStamps_tes[pid]
            isRealBGL[pid] += isRealBGL_tes[pid]
            args.test_portion[pid] = len(timeStamps_tes[pid])

    if args.mask_interpolated:
        _realBGLMsk = isRealBGL
    else:
        _realBGLMsk = None

    ## open a mask file and feed it to prepare_for_learning if necessary
    #  with open('uva_whatIfMask_60min.pkl', 'rb') as f:
    #    otherMask = pickle.load(f)

    data, _ = prepare_for_learning(dataRaw, timeStamps, _realBGLMsk, None, args.seq_len, args.batch_size,
                                args.pred_horizon, krs_backend.floatx(), args.bgl_scale, args.scale_for_others,
                                testPortion=args.test_portion, validPortion=args.valid_portion,
                                normalize=args.normalize, stateful=0, useLabelDiff=args.label_differencing,
                                filterWhatIfPoints=args.filter_whatif, filterPreBreak=False,
                                whatIfFeatures=[args.cats_to_load.index(x) for x in ['meal', 'bolus'] if x in args.cats_to_load],
                                whatIfScenario=args.whatif_scenario)
    # data[pID] will be be in N x SEQ_LEN x D shape, labels in N x 1 shape

profiles = get_profiles(args.mode, data, args.pid_list, args.pred_horizon, args.early_stop)


#### KERAS MODEL ####
dims = {'seq_len': args.seq_len, 'D': data[args.pid_list[0]]['train'].shape[2], 'pred_horizon': args.pred_horizon}
args.gpu_id = 0
args.gpu_mem = 0.05

if args.whatif_scenario:
    print('\nBuilding and compiling a double-LSTM Keras model...')
    if args.double_output_model:
        print("  With double output...\n")
    model = DoubleLstm(args, dims, double_output=args.double_output_model,
                       whatif_inds=[args.cats_to_load.index(x) for x in ['meal', 'bolus'] if x in args.cats_to_load])
else:
    print('\nBuilding and compiling a simple LSTM Keras model...')
    model = KerasSimpleRnn(args, dims)

model.generate()


#### TRAINING & EVALUATION ####
test_costHist = []
test_predHist = []
t0 = time.time()
initialWeights = model.internal_model.get_weights()
for profile in profiles:
    print('\nDoing profile: {}'.format(profile))

    if isinstance(profile, str):
        if profile.startswith('SAVE_WEIGHTS'):
            _fname = profile.partition(' to ')[2]
            savedWs = model.internal_model.get_weights()
            if _fname:
                np.save(_fname, savedWs)
        elif profile.startswith('RESTORE_WEIGHTS'):
            _fname = profile.partition(' from ')[2]
            if _fname:
                model.internal_model.set_weights(np.load(_fname))
            else:
                model.internal_model.set_weights(savedWs)
        elif profile == 'RESET_WEIGHTS':
            rndWeights = [np.random.permutation(w.flat).reshape(w.shape) for w in initialWeights]
            rndWeights[2] = initialWeights[2]  # maintain the biases
            model.internal_model.set_weights(rndWeights)
    else:
        curData = mix_data(data, profile, AxisOfN=0)
        q = model.seq_len

        if 'train' in curData:
            for init_lrate, max_epochs in profile['LRATE']:
                model.compile(init_lrate=init_lrate, lrate_decay=args.lrate_decay)

                _valData = None
                _valPortion = 0.0
                if 'validate' in curData:
                    _valData = [curData['validate'], curData['validateLabels']]
                elif 'VALIDATE_ON' in profile:
                    _valPortion = profile['VALIDATE_ON']

                if 'EARLY_STOP' in profile:
                    early_stop = profile['EARLY_STOP']
                else:
                    print("Warning: no early stop provided in this profile.")
                    early_stop = None

                print("\n> Going for {} epochs with lrate {}...".format(max_epochs, init_lrate))
                fit_hist = model.fit(curData['train'], curData['trainLabels'],
                                     val_data=_valData, val_portion=_valPortion, batch_size=args.batch_size,
                                     early_stop=early_stop, max_epochs=max_epochs)

        # Evaluation
        if 'test' in curData:
            print("\nEvaluating on test data...")
            test_costHist.append(model.evaluate(curData['test'], curData['testLabels'], args.batch_size))
            test_predHist.append(model.predict(curData['test'], args.batch_size))
            if args.mode == 'icmla200':
                print('Actual value: {}, predicted: {}, cost: {}, RMSE so far: {}'
                      .format(curData['testLabels'], test_predHist[-1], test_costHist[-1],
                              np.mean(test_costHist) ** 0.5))
            else:
                _curAbsolutePreds = test_predHist[-1][:, 0]
                _curAbsoluteTargets = curData['testLabels'][:, 0]
                if args.label_differencing:
                    _curAbsolutePreds += curData['test'][:, -1, 0]
                    _curAbsoluteTargets += curData['test'][:, -1, 0]
                print('test_costHist: {}, average RMSE: {}, hypo time delays: {}, hyper time delays: {}'
                      .format(test_costHist,
                              np.mean([x ** 0.5 for x in test_costHist]) / args.bgl_scale,
                              bgl_time_delay(_curAbsolutePreds, _curAbsoluteTargets, 70 * args.bgl_scale, lambda a, b: a <= b),
                              bgl_time_delay(_curAbsolutePreds, _curAbsoluteTargets, 130 * args.bgl_scale, lambda a, b: a >= b),
                              )
                      )

print('\nTotal time: {}'.format(time.time() - t0))
print('\nArgs: ', args)

if args.result_output_file is not None:
    with open(args.result_output_file, 'a') as f:
        if args.mode == 'icmla200':
            _rmse = np.mean(test_costHist) ** 0.5
        else:
            _rmse = np.mean([x ** 0.5 for x in test_costHist])
        f.write('test_costHist: {}, average RMSE: {}'
                .format(test_costHist, _rmse / args.bgl_scale))
        f.write('\nTotal time: {}\n\n'.format(time.time() - t0))
        #    for arg in args._get_kwargs():
        #      f.write('{}\t\t: {}\n'.format(arg[0], arg[1]))

        f.write('\n-----------------------\n')
