### This tensorflow script aims at training and evaluating an RNN on the BGL time
### series data loaded from various sources.

### Sadegh M, Summer 2017

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy as np
import tensorflow as tf
import argparse
import time
import os
from patient_data import prepare_for_learning, load_from_csv, load_from_xml, load_from_KalmanPatientXML
from patient_data import mix_data, concatData, prepadData, add_extra_feature
from nn_models_tf import RNNTSAModel, Optimizer
from shared_args import parse_cmd_arguments, customize_args, get_profiles


args = parse_cmd_arguments( argparse.ArgumentParser() )

if args.random_seed is not None:
  np.random.seed(args.random_seed)

args.seq_len = args.seq_len_in_minutes // args.data_resolution
args.pred_horizon = args.pred_horizon_in_minutes // args.data_resolution

args = customize_args(args)

######## PREPARATION ########
if args.data_pkl != '':
  print('\nLoading data from pickle file {}\n'.format(args.data_pkl))
  import pickle
  with open(args.data_pkl, 'rb') as f:
    data = pickle.load(f)
elif 'data' not in globals():
  print('Loading data for patients ', args.pid_list)

  if args.mode == 'synthetic':
    datapXML, timeStamps, isRealBGL = load_from_csv(zip(args.pid_list, args.data_files), ' ',
                                                    args.data_resolution, True, True,
                                                    args.cats_to_load)
  else:
    datapXML, timeStamps, isRealBGL = load_from_xml(zip(args.pid_list, args.data_files), args.data_resolution, args.cats_to_load, verbose=False)

  if args.data_files_kp and args.cats_to_load_kp:
    datakpXML = load_from_KalmanPatientXML(zip(args.pid_list, args.data_files_kp), timeStamps, args.cats_to_load_kp)
    dataRaw = concatData(args.pid_list, (datapXML, datakpXML), axis=1)
  else:
    datakpXML = []
    dataRaw = datapXML

  if args.prepad_dic:
    print('\nPrepadding the loaded raw data using given sources...')
    dataRaw, timeStamps, isRealBGL = prepadData(dataRaw, isRealBGL, timeStamps, args.prepad_dic, prepadLength=args.seq_len)

  dataRaw = add_extra_feature(dataRaw, args.other_cats_to_add)
  _both_cats = args.cats_to_load+args.cats_to_load_kp

  if args.mask_interpolated:
    labelMask = isRealBGL
  else:
    labelMask = None

  # data[pID] will be be in N x SEQ_LEN x D shape, labels in N x 1 shape
  data, _ = prepare_for_learning(dataRaw, timeStamps, labelMask, args.seq_len, args.batch_size,
                              args.pred_horizon, 'float32', args.bgl_scale, args.other_scale,
                              testPortion=args.test_portion, validPortion=args.valid_portion,
                              normalize=args.other_normalize, stateful=0, useLabelDiff=args.label_differencing,
                              filterWhatIfPoints=args.filter_whatif, filterPreBreak=False,
                              whatIfFeatures=[_both_cats.index(x) for x in ['meal', 'bolus'] if x in _both_cats])

if data[args.pid_list[0]]['all'] is not None:
  N, T, D = data[args.pid_list[0]]['all'].shape
else:
  raise Exception("Loaded 'data' is empty or not properly structured.")

profiles = get_profiles(args.mode, data, args.pid_list, args.pred_horizon, early_stop=args.early_stop)

gpu_options = tf.GPUOptions(per_process_gpu_memory_fraction=0.05, allow_growth=False)
gpu_config = tf.ConfigProto(log_device_placement=False, gpu_options=gpu_options)
with tf.Graph().as_default(), tf.Session(config=gpu_config) as sess:
  if args.random_seed is not None:
    tf.set_random_seed(args.random_seed)
#  data_ts               = tf.placeholder(tf.float32, shape=[None, T, D]) # our data is not time-major
#  targets_ts            = tf.placeholder(tf.float32, shape=[None, 1])
#  initial_lrate_ts      = tf.placeholder(dtype=tf.float32)
#
#  with tf.variable_scope("tsa_model", reuse=None):
#    tsa_model = RNNTSAModel(args, data_ts)
#
#  with tf.variable_scope("tsa_optimizer", reuse=None):
#    tsa_optimizer = Optimizer(data_ts, tsa_model.reg_output, targets_ts, initial_lrate_ts, 'mse')
#
#  tsa_optimizer.compile_train_op(args.optimizer, tf.trainable_variables(), args.grad_clip)

#  saver_global = tf.train.Saver(tf.global_variables(), max_to_keep=100)
#  saver_tsa_model = tf.train.Saver(var_list=tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope='tsa_model'), max_to_keep=100)

  print("Training started...")
#  tf.global_variables_initializer().run()
  profile_cost_hist = []
  all_jobs_start_time = time.time()

  for profile in profiles:
    print('\nDoing profile: {}'.format(profile))
    if isinstance(profile, str):
      if profile.startswith('SAVE_WEIGHTS'):
        _fname = profile.partition(' to ')[2]
        if _fname:
          saver_global.save(sess, os.path.join(args.save_dir, _fname), global_step=None)
        else:
          saver_global.save(sess, os.path.join(args.save_dir, 'savedWs'), global_step=None)
      elif profile.startswith('RESTORE_WEIGHTS'):
        _fname = profile.partition(' from ')[2]
        if _fname:
          saver_global.restore(sess, os.path.join(args.save_dir, _fname))
        else:
          saver_global.restore(sess, os.path.join(args.save_dir, 'savedWs'))
      elif profile == 'RESET_WEIGHTS':
        tf.global_variables_initializer().run()
    else:
      if 'BTCH_SZ' in profile:
        _btchsz = profile['BTCH_SZ']
      else:
        _btchsz = args.batch_size

      if 'SHUFFLE' in profile:
        _shuffle = profile['SHUFFLE']
      else:
        _shuffle = True

      if 'EARLY_STOP' not in profile:
        early_stop_thresh = 0
      else:
        early_stop_thresh = profile['EARLY_STOP']

      curData = mix_data(data, profile, AxisOfN=0, shuffle=_shuffle)

      data_ts               = tf.placeholder(tf.float32, shape=[None, T, D]) # our data is not time-major
      targets_ts            = tf.placeholder(tf.float32, shape=[None, 1])
      initial_lrate_ts      = tf.placeholder(dtype=tf.float32)

      with tf.variable_scope("tsa_model", reuse=False):
        tsa_model = RNNTSAModel(args, data_ts)

      with tf.variable_scope("tsa_optimizer", reuse=False):
        tsa_optimizer = Optimizer(data_ts, tsa_model.reg_output, targets_ts, initial_lrate_ts, 'mse')

      tsa_optimizer.compile_train_op(args.optimizer, tf.trainable_variables(), args.grad_clip)

      saver_global = tf.train.Saver(max_to_keep=1000)
      #saver_tsa_model = tf.train.Saver(var_list=tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope='tsa_model'), max_to_keep=1000)

      tf.global_variables_initializer().run()

      if 'train' in curData:
        min_val_loss = float('inf')
        for lrate, nbeps in profile['LRATE']:
          print("\n> Going for {} epochs with initial lrate {}, batch size {}, n_train {} and n_valid {}..."
                    .format(nbeps, lrate, _btchsz, curData['train'].shape[0], curData['validate'].shape[0]))

          e = 0
          epoch_val_cost_hist = []
          while e < nbeps:
            e += 1
            t0 = time.time()
            train_results = tsa_optimizer.go_over_data(curData['train'], curData['trainLabels'],
                                                  _btchsz, sess, train=True,
                                                  feed_dict={tsa_model.output_dropout_ts:tsa_model.output_dropout, initial_lrate_ts:lrate})

            val_results = tsa_optimizer.go_over_data(curData['validate'], curData['validateLabels'],
                                                  _btchsz, sess, train=False,
                                                  feed_dict={tsa_model.output_dropout_ts:0.0},
                                                  save_in_hist=False)

            epoch_val_cost_hist.append(val_results['mean_loss'])

            print("Epoch {:3d} | Loss {:.7f} | val_loss {:.7f} | lr: {:.5f} | {:.3f} sec"
                    .format(e, train_results['mean_loss'], val_results['mean_loss'], lrate, time.time()-t0))

            if val_results['mean_loss'] < min_val_loss:
              min_val_loss = val_results['mean_loss']
              print("Saving best weights...")
              saver_global.save(sess, os.path.join(args.save_dir, 'best_val_weights'), global_step=None)

            elif early_stop_thresh > 0 and len(epoch_val_cost_hist) > early_stop_thresh:
              if min_val_loss != np.min(epoch_val_cost_hist[-early_stop_thresh:]):
                print("Restoring best weights saved with loss ", min_val_loss)
                saver_global.restore(sess, os.path.join(args.save_dir, 'best_val_weights'))
                break

      if 'test' in curData:
        print("\nEvaluating on test data...")
        test_results = tsa_optimizer.go_over_data(curData['test'], curData['testLabels'],
                                              _btchsz, sess, train=False,
                                              feed_dict={tsa_model.output_dropout_ts:0.0},
                                              save_in_hist=False)
        profile_cost_hist.append(test_results['mean_loss'])
        print('Test cost history: {}'.format(profile_cost_hist))

  print("\nAll jobs done in {:5f} seconds. Mean RMSE: {}".format(
                          time.time() - all_jobs_start_time,
                          np.mean([x**0.5 for x in profile_cost_hist]) / args.bgl_scale))

print('\nArgs: ', args)