### Unifies the argument reading that is shared between various main learning codes.

from patient_data import get_200_indices


def parse_cmd_arguments(parser):
    """
    given a parser of the type argparse.ArgumentParser(), adds arguments and
    returns the args object. This function is used to share arguments.
    """
    parser.add_argument('--save_dir', type=str, default='/home/sadegh/diabetes/Code/tf_save',
                        help="directory to save to and restore models from")

    parser.add_argument('--result_output_file', type=str, default=None,
                        help="file to save the results in")

    parser.add_argument('--data_pkl', type=str, default='',
                        help="if provided, loads this pkl file and moves on, otherwise reads from files")

    parser.add_argument('--mode', type=str, default='workshop-basis',
                        help="""MODE options are 'icmla200', 'workshop-basis'', 'synthetic' and 'basis'. Each mode
                              determines what data are loaded, what profiles are created, etc. 'normal' can
                              be used for testing. Other modes are for special cases.""")

    parser.add_argument('--random_seed', type=int, default=None,
                        help='if not None, the seed will be set to this value')

    parser.add_argument('--mask_interpolated', type=bool, default=True,
                        help='mask the part of data that is interpolated (do not use it as labels for learning)')

    parser.add_argument('--label_differencing', type=bool, default=True,
                        help='if True delta_bgl is used as the label for training and testing')

    parser.add_argument('--filter_whatif', type=bool, default=False,
                        help='if True, filter out data points with what if events between t0 and t0+pred_horizon')

    parser.add_argument('--whatif_scenario', type=bool, default=False,
                        help='if True, trains with double RNN in the whatif scenario')

    parser.add_argument('--double_output_model', type=bool, default=False,
                        help='double output model has 2 losses')

    parser.add_argument('--cats_to_load', type=str, default="['bgl', 'meal', 'bolus']",
                        # 'basis_gsr', 'basis_heart_rate', 'basis_skin_temperature']",
                        help='categories to load from xml files')

    parser.add_argument('--other_cats_to_add', type=str, default="[]",
                        help='other cats like "zero" or "random" or "retrospect_?" to add to the features')

    parser.add_argument('--layer_size', type=int, default=20,
                        help='size of RNN hidden state')

    parser.add_argument('--num_layers', type=int, default=1,
                        help='number of RNN layers in the model, each will have the same size')

    parser.add_argument('--cell_type', type=str, default='lstm',
                        help='one of rnn, gru, or lstm')

    parser.add_argument('--kernel_activation', type=str, default='tanh',
                        help='the kernel activation (not the gates). supported are tanh, sigmoid, relu and softplus')

    parser.add_argument('--batch_size', type=int, default=512,
                        help='minibatch size')

    parser.add_argument('--seq_len_in_minutes', type=int, default=6 * 60,
                        help='sequence length in minutes (data resolution determines the length that model sees)')

    parser.add_argument('--data_resolution', type=int, default=5,
                        help='resolution of data in minutes')

    parser.add_argument('--pred_horizon_in_minutes', type=int, default=60,
                        help='prediction horizon in minutes')

    parser.add_argument('--valid_portion', type=int, default=10 * 24 * 12,
                        help='the number of HOURS used for validation')

    parser.add_argument('--test_portion', type=int, default=10 * 24 * 12,
                        help='the number of HOURS used for testing')

    parser.add_argument('--bgl_scale', type=float, default=1 / 600,
                        help='scale specific to the bgl values')

    parser.add_argument('--scale_for_others', type=float, default=1 / 3,
                        help='scale applied to other features (after normalization if any)')

    parser.add_argument('--normalize', type=bool, default=True,
                        help='normalization applied to features other than BGL')

    parser.add_argument('--grad_clip', type=float, default=None,
                        help='maximum permissible norm of the gradient')

    parser.add_argument('--lrate_decay', type=float, default=0.0,
                        help='the decay of the learning rate, 0.0 means no decay')

    parser.add_argument('--early_stop', type=int, default=2,
                        help='do validation early stopping with this threshold')

    parser.add_argument('--output_dropout', type=float, default=0.0,
                        help='the probability of dropping outgoing weights for each layer')

    parser.add_argument('--optimizer', type=str, default='adam',
                        help='one of sgd, momentum, rmsprop, or adam')

    parser.add_argument('--kernel_initializer', type=str, default='glorot_uniform',
                        help='initializer used for non-recursive weights')

    return parser.parse_args()


def customize_args(args):
    """
  takes the args dictionary and customizes it based on 'mode', then returns the result.
  """

    """ 'prepad_dic' is a dictionary in the form p1:(p2, idx) where we say for patient p1,
  prepad it with data from patient p2 ending at index idx (take from idx and before)
  The prepadding length is args.seq_len-1 which is the maximum required. """
    args.prepad_dic = None

    if args.mode == 'icmla200':
        args.pid_list = ['304', '306', '310', '312', '316', '302', '308', '309', '313', '307']
        args.data_files = ['../Data/p' + pID + '.xml' for pID in args.pid_list]
        args.prepad_dic = {'312': ('306', 7606), '304': ('316', 23208), '306': ('312', 19450),
                           '310': ('312', 20697), '316': ('304', 15973)}

    elif args.mode == 'basis':
        args.pid_list = ['501', '502', '503', '504', '505', '506']
        args.data_files = ['/home/sadegh/Data/op' + pID + '.xml' for pID in args.pid_list]

    elif args.mode == 'workshop-basis':
        args.pid_list = ['559', '563', '570', '575', '588', '591']
        args.data_files = ['/home/sadegh/diabetes/Data/' + pID + '-ws-training.xml' for pID in args.pid_list]
        args.data_files_test = ['/home/sadegh/diabetes/Data/' + pID + '-ws-testing.xml' for pID in args.pid_list]


    elif args.mode == 'synthetic':
        args.pid_list = ['{}_adult#{:03d}'.format(x, x) for x in range(5, 10)]
        args.data_files = ['/home/sadegh/diabetes/Data/T1DMS/{}.csv'.format(x) for x in args.pid_list]

    else:
        raise Exception('Mode is not recognized : ', args.mode)

    return args


def get_profiles(mode, data, pid_list, pred_horizon, early_stop=None):
    """ 'profiles' is a list. Each of its items is a profile, usually a dictionary.
        The keys TRAIN_ON, VALIDATE_ON and TEST_ON specifcy the data for each purpose.
        'TRAIN_ON': {'312':'train', '310':'all', '304':(100,None)} means train on the
        training portion of 312, all of 310, and from 100 to the end of 304.
        Each profile has also the key 'LRATE', which is a list of pairs of (lrate, epochs).
        Each profile can optionally have 'BTCH_SZ' which will override the global BTCH_SZ param.
        Special profiles 'SAVE_WEIGHTS' 'RESTORE_WEIGHTS' and 'RESET_WEIGHTS' are used to save
        the weights, restore the last saved weights and reset to initial weights respectively.
        These are not dictionaries, just texts you put in 'profiles'. """
    profiles = []

    print("\nMODE is '{}': creating profiles...".format(mode))

    if mode == 'icmla200':
        _tsFileNames = [(pID, '../Data/' + pID + '.ts') for pID in pid_list[:5]]
        indices = get_200_indices(data, _tsFileNames)
        for pID in pid_list[:5]:
            profiles.append({'TRAIN_ON': {p: 'all' for p in pid_list[:-2] if p != pID},
                             'VALIDATE_ON': {p: 'all' for p in pid_list[-2:]},
                             'TEST_ON': None,
                             'LRATE': [(0.01, 500), (0.01, 500)],
                             'EARLY_STOP': early_stop
                             })
            profiles.append('SAVE_WEIGHTS')
            for i in indices[pID]:
                if i > pred_horizon:
                    print("Making profile for index {} of patient {}...".format(i, pID))
                    _teston = {pID: (i, i + 1)}
                    _m = max(indices[pID])
                    profiles.append({'TRAIN_ON': {pID: (0, i - pred_horizon + 1)},
                                     'TEST_ON': _teston,
                                     'VALIDATE_ON': {p: 'all' for p in pid_list[-2:]},
                                     # 'LRATE': [(0.001, int(5 + 15 * i / _m)), (0.0001, int(1 + 15 * i / _m))]
                                     'LRATE': [(0.001, 500), (0.0001, 500)],
                                     'EARLY_STOP': early_stop
                                     })
                profiles.append('RESTORE_WEIGHTS')
            profiles.append('RESET_WEIGHTS')

    elif mode == 'basis' or mode == 'workshop-basis':
        profiles.append({'TRAIN_ON': {p: 'train' for p in pid_list},
                         'VALIDATE_ON': {p: 'validate' for p in pid_list},
                         'TEST_ON': None,
                         'LRATE': [(0.01, 100), (0.001, 100)],
                         'EARLY_STOP': early_stop
                         })
        profiles.append('SAVE_WEIGHTS')
        for pID in pid_list:
            #        profiles.append({'TRAIN_ON':{p:'all' for p in pid_list if p!=pID},
            #                         'VALIDATE_ON': {pID:'validate'},
            #                         'TEST_ON':     None,
            #                         'LRATE':       [(0.01, 500), (0.001, 500)],
            #                         'EARLY_STOP':  early_stop
            #                         } )
            profiles.append({'TRAIN_ON': {pID: 'train'},
                             'TEST_ON': {pID: 'test'},
                             'VALIDATE_ON': {pID: 'validate'},
                             'LRATE': [(0.001, 100), (0.0001, 100)],
                             'EARLY_STOP': early_stop})
            profiles.append('RESTORE_WEIGHTS')

    elif mode == 'normal':
        pID = '501'
        profiles.append({'TRAIN_ON': {p: 'all' for p in pid_list if p != pID},
                         'TEST_ON': {pID: 'test'},
                         'VALIDATE_ON': {pID: 'validate'},
                         'LRATE': [(0.001, 500)],
                         'EARLY_STOP': early_stop})
    #        profiles.append('RESET_WEIGHTS')

    elif mode == 'synthetic':
        for pID in pid_list:
            profiles.append({'TRAIN_ON': {pID: 'train'},
                             'TEST_ON': {pID: 'test'},
                             'VALIDATE_ON': {pID: 'validate'},
                             'LRATE': [(0.01, 500), (0.001, 500)],
                             'EARLY_STOP': early_stop})
            profiles.append('RESET_WEIGHTS')

    return profiles
